// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.io.*;
import java.net.*;
import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;
import org.bouncycastle.util.encoders.Base64;

/**
 *  An XML-RPC client implemented as a JavaScript Host Object. See the
 *  XWT spec for information on its behavior.
 *
 *  NOTE: this client is EXTREMELY lenient in the responses it will
 *  accept; there are many, many invalid responses that it will
 *  successfully parse and return. Do NOT use this to determine the
 *  validity of your server.
 *
 *  This client conforms to <a href="http://www.xmlrpc.com/spec">The
 *  XML-RPC Spec</a>, subject to these limitations:
 *  <ol>
 *    <li> XMLRPC cannot invoke methods that require a <base64/> argument
 *    <li> if a return value contains a <base64/>, it will be returned as a string
 *    <li> The decision to pass a number as <i4/> or <double/> is based
 *         entirely on whether or not the argument is fractional. Thus, it
 *         is impossible to pass a non-fractional number to an xmlrpc
 *         method that insists on being called with a <double/> element. We
 *         hope that most xml-rpc servers will be able to automatically
 *         convert.
 *  </ol>
 */
class XMLRPC extends JS.Callable {

    public Object[] keys() { throw new Error("not implemented"); }

    /** the url to connect to */
    protected String url = null;

    /** the method name to invoke on the remove server */
    protected String methodname = null;

    /** this holds character content as we read it in -- since there is only one per instance, we don't support mixed content */
    protected AccessibleCharArrayWriter content = new AccessibleCharArrayWriter(100);

    /** the HTTP connection to use */
    protected HTTP http = null;

    /** The object stack. As we process xml elements, pieces of the
     *  return value are pushed onto and popped off of this stack.
     *
     *  The general protocol is that any time a &lt;value&gt; tag is
     *  encountered, an empty String ("") is pushed onto the stack. If
     *  the &lt;value/&gt; node has content (either an anonymous
     *  string or some other XML node), that content replaces the
     *  empty string.
     *
     *  If an &lt;array&gt; tag is encountered, a null is pushed onto the
     *  stack. When a &lt;/data&gt; is encountered, we search back on the
     *  stack to the last null, replace it with a NativeJS.Array, and
     *  insert into it all elements above it on the stack.
     *
     *  If a &lt;struct&gt; tag is encountered, a JSObject is pushed
     *  onto the stack. If a &lt;name&gt; tag is encountered, its CDATA is
     *  pushed onto the stack. When a &lt;/member&gt; is encountered, the
     *  name (second element on stack) and value (top of stack) are
     *  popped off the stack and inserted into the struct (third
     *  element on stack).
     */
    protected Vec objects = null;

    /** used to detect multi-ref data */
    private Hash tracker;

    /** True iff the return value is a fault (and should be thrown as an exception) */
    protected boolean fault = false;


    // Methods to Recieve and parse XML-RPC Response ////////////////////////////////////////////////////

    private class Helper extends XML {
        public Helper() { super(BUFFER_SIZE); }

        public void startElement(XML.Element c) {
            content.reset();
            if (c.localName.equals("fault")) fault = true;
            else if (c.localName.equals("struct")) objects.setElementAt(new JS.Obj(), objects.size() - 1);
            else if (c.localName.equals("array")) objects.setElementAt(null, objects.size() - 1);
            else if (c.localName.equals("value")) objects.addElement("");
        }
        
        public void endElement(XML.Element c) {
            
            if (c.localName.equals("int") || c.localName.equals("i4"))
                objects.setElementAt(new Integer(new String(content.getBuf(), 0, content.size())), objects.size() - 1);
            
            else if (c.localName.equals("boolean"))
                objects.setElementAt(content.getBuf()[0] == '1' ? Boolean.TRUE : Boolean.FALSE, objects.size() - 1);
            
            else if (c.localName.equals("string"))
                objects.setElementAt(new String(content.getBuf(), 0, content.size()), objects.size() - 1);
            
            else if (c.localName.equals("double"))
                objects.setElementAt(new Double(new String(content.getBuf(), 0, content.size())), objects.size() - 1);
            
            else if (c.localName.equals("base64"))
                objects.setElementAt(new ByteStream(Base64.decode(new String(content.getBuf(), 0, content.size()))), objects.size() - 1);
            
            else if (c.localName.equals("name"))
                objects.addElement(new String(content.getBuf(), 0, content.size()));
            
            else if (c.localName.equals("value") && "".equals(objects.lastElement()))
                objects.setElementAt(new String(content.getBuf(), 0, content.size()), objects.size() - 1);
            
            else if (c.localName.equals("dateTime.iso8601")) {
                throw new Error("not implemented");
                /*
                  String s = new String(content.getBuf(), 0, content.size());
                  
                  // strip whitespace
                  int i=0;
                  while(Character.isWhitespace(s.charAt(i))) i++;
                  if (i > 0) s = s.substring(i);
                  
                  try {
                  NativeDate nd = (NativeDate)JS.Thread.enter().newObject(org.xwt.util.JS.Obj.defaultObjects, "Date");
                  double date = NativeDate.date_msecFromDate(Double.valueOf(s.substring(0, 4)).doubleValue(),
                  Double.valueOf(s.substring(4, 6)).doubleValue() - 1,
                  Double.valueOf(s.substring(6, 8)).doubleValue(),
                  Double.valueOf(s.substring(9, 11)).doubleValue(),
                  Double.valueOf(s.substring(12, 14)).doubleValue(),
                  Double.valueOf(s.substring(15, 17)).doubleValue(),
                  (double)0
                  );
                  nd.jsFunction_setTime(NativeDate.internalUTC(date));
                  objects.setElementAt(nd, objects.size() - 1);
                  
                  } catch (Exception e) {
                  if (Log.on) Log.log(this, "error parsing date : " + s);
                  if (Log.on) Log.log(this, e);
                  }
                */
                
            } else if (c.localName.equals("member")) {
                Object memberValue = objects.elementAt(objects.size() - 1);
                String memberName = (String)objects.elementAt(objects.size() - 2);
                JS struct = (JS)objects.elementAt(objects.size() - 3);
                struct.put(memberName, memberValue);
                objects.setSize(objects.size() - 2);
                
            } else if (c.localName.equals("data")) {
                int i;
                for(i=objects.size() - 1; objects.elementAt(i) != null; i--);
                JS.Array arr = new JS.Array();
                for(int j = i + 1; j<objects.size(); j++) arr.put(new Integer(j - i - 1), objects.elementAt(j));
                objects.setElementAt(arr, i);
                objects.setSize(i + 1);
                
            }
            
            content.reset();
        }
        
        public void characters(char[] ch, int start, int length) {
            try { content.write(ch, start, length); }
            catch (Exception e) { 
                if (Log.on) Log.log(this, "Exception in XMLRPC.content() -- this should never happen");
                if (Log.on) Log.log(this, e);
            }
        }
        
        public void whitespace(char[] ch, int start, int length) {}
    }

    // Methods to make outbound XML-RPC request ///////////////////////////////////////////////////

    /** Appends the XML-RPC representation of <code>o</code> to <code>sb</code> */
    void appendObject(Object o, StringBuffer sb) throws JS.Exn {

        if (o == null) {
            throw new JS.Exn("attempted to send a null value via XML-RPC");

        } else if (o instanceof Number) {
            if ((double)((Number)o).intValue() == ((Number)o).doubleValue()) {
                sb.append("                <value><i4>");
                sb.append(((Number)o).intValue());
                sb.append("</i4></value>\n");
            } else {
                sb.append("                <value><double>");
                sb.append(o);
                sb.append("</double></value>\n");
            }

        } else if (o instanceof Boolean) {
            sb.append("                <value><boolean>");
            sb.append(((Boolean)o).booleanValue() ? "1" : "0");
            sb.append("</boolean></value>\n");

        } else if (o instanceof ByteStream) {
            try {
                sb.append("                <value><base64>\n");
                InputStream is = ((ByteStream)o).getInputStream();
                byte[] buf = new byte[54];
                while(true) {
                    int numread = is.read(buf, 0, 54);
                    if (numread == -1) break;
                    byte[] writebuf = buf;
                    if (numread < buf.length) {
                        writebuf = new byte[numread];
                        System.arraycopy(buf, 0, writebuf, 0, numread);
                    }
                    sb.append("              ");
                    sb.append(new String(Base64.encode(writebuf)));
                    sb.append("\n");
                }
                sb.append("\n              </base64></value>\n");
            } catch (IOException e) {
                if (Log.on) Log.log(this, "caught IOException while attempting to send a ByteStream via XML-RPC");
                if (Log.on) Log.log(this, e);
                throw new JS.Exn("caught IOException while attempting to send a ByteStream via XML-RPC");
            }

        } else if (o instanceof String) {
            sb.append("                <value><string>");
            String s = (String)o;
            if (s.indexOf('<') == -1 && s.indexOf('&') == -1) {
                sb.append(s);
            } else {
                char[] cbuf = s.toCharArray();
                int oldi = 0, i=0;
                while(true) {
                    while(i < cbuf.length && cbuf[i] != '<' && cbuf[i] != '&') i++;
                    sb.append(cbuf, oldi, i - oldi);
                    if (i >= cbuf.length) break;
                    if (cbuf[i] == '<') sb.append("&lt;");
                    else if (cbuf[i] == '&') sb.append("&amp;");
                    i = oldi = i + 1;
                    if (i >= cbuf.length) break;
                }
            }
            sb.append("</string></value>\n");

            /*
        } else if (o instanceof org.xwt.js.Date) {
            sb.append("                <value><dateTime.iso8601>");
              FIXME
            org.xwt.js.Date d = (org.xwt.js.Date)o;
            Date d = new Date(nd.getRawTime());
            sb.append(d.getYear() + 1900);
            if (d.getMonth() + 1 < 10) sb.append('0');
            sb.append(d.getMonth() + 1);
            if (d.getDate() < 10) sb.append('0');
            sb.append(d.getDate());
            sb.append('T');
            if (d.getHours() < 10) sb.append('0');
            sb.append(d.getHours());
            sb.append(':');
            if (d.getMinutes() < 10) sb.append('0');
            sb.append(d.getMinutes());
            sb.append(':');
            if (d.getSeconds() < 10) sb.append('0');
            sb.append(d.getSeconds());
            sb.append("</dateTime.iso8601></value>\n");
            */

        } else if (o instanceof JS.Array) {
            if (tracker.get(o) != null) throw new JS.Exn("attempted to send multi-ref data structure via XML-RPC");
            tracker.put(o, Boolean.TRUE);
            sb.append("                <value><array><data>\n");
            JS.Array a = (JS.Array)o;
            for(int i=0; i<a.length(); i++) appendObject(a.elementAt(i), sb);
            sb.append("                </data></array></value>\n");

        } else if (o instanceof JS) {
            if (tracker.get(o) != null) throw new JS.Exn("attempted to send multi-ref data structure via XML-RPC");
            tracker.put(o, Boolean.TRUE);
            JS j = (JS)o;
            sb.append("                <value><struct>\n");
            Object[] ids = j.keys();
            for(int i=0; i<ids.length; i++) {
                sb.append("                <member><name>" + ids[i] + "</name>\n");
                appendObject(j.get(ids[i].toString()), sb);
                sb.append("                </member>\n");
            }
            sb.append("                </struct></value>\n");

        } else {
            throw new JS.Exn("attempt to send object of type " + o.getClass().getName() + " via XML-RPC");

        }
    }

    // this is synchronized in case multiple threads try to make a call on the same object... in the future, change this
    // behavior to use pipelining.
    public synchronized Object call2(JS.Array args) throws JS.Exn, IOException {
        if (Log.verbose) Log.log(this, "call to " + url + " : " + methodname);

        if (tracker == null) tracker = new Hash();
        else tracker.clear();

        if (objects == null) objects = new Vec();
        else objects.setSize(0);

        String content = send(args, http);
        if (Log.verbose) {
            String s;
            BufferedReader br2 = new BufferedReader(new StringReader(content));
            while ((s = br2.readLine()) != null) Log.log(this, "send: " + s);
        }

        HTTP.HTTPInputStream is = http.POST("text/xml", content);
        try {
            BufferedReader br = !Log.verbose ?
                new BufferedReader(new InputStreamReader(new Filter(is))) :
                new BufferedReader(new FilterReader(new InputStreamReader(new Filter(is))) {
                        public int read() throws IOException {
                            int i = super.read();
                            if (Log.on) Log.log(this, "recv: " + ((char)i));
                            return i;
                        }
                        public int read(char[] c, int off, int len) throws IOException {
                            int ret = super.read(c, off, len);
                            if (ret == -1) return ret;
                            String s;
                            BufferedReader br2 = new BufferedReader(new StringReader(new String(c, off, ret)));
                            while ((s = br2.readLine()) != null) Log.log(this, "recv: " + s);
                            return ret;
                        }
                    });
            return recieve(br);
        } finally {
            is.close();
        }
    }

    protected String send(JS.Array args, HTTP http) throws JS.Exn, IOException {
        StringBuffer content = new StringBuffer();
        content.append("\r\n");
        content.append("<?xml version=\"1.0\"?>\n");
        content.append("    <methodCall>\n");
        content.append("        <methodName>");
        content.append(methodname);
        content.append("</methodName>\n");
        content.append("        <params>\n");
        for(int i=0; i<args.length(); i++) {
            content.append("            <param>\n");
            appendObject(args.elementAt(i), content);
            content.append("            </param>\n");
        }
        content.append("        </params>\n");
        content.append("    </methodCall>");
        return content.toString();
    }
        
    protected Object recieve(BufferedReader br) throws JS.Exn, IOException {
        // parse XML reply
        try {
            new Helper().parse(br);
        } catch (XML.XMLException e) {
            if (Log.on) Log.log(this, "reply from server was not well-formed XML: " + e);
            throw new JS.Exn("reply from server was not well-formed XML: " + e);
        }
        
        if (fault) throw new JS.Exn(objects.elementAt(0));
        if (objects.size() == 0) return null;
        return objects.elementAt(0);
    }

    public final Object call(JS.Array args) throws JS.Exn {

        if (!ThreadMessage.suspendThread()) return null;

        try {
            return call2(args);
        } catch (IOException se) {
            if (Log.on) Log.log(this, se);
            throw new JS.Exn("socket exception: " + se);

        } catch (JS.Exn jse) {
            if (Log.on) Log.log(this, jse.toString());
            throw jse;
        } finally {
            ThreadMessage.resumeThread();
        }

    }

    /** When you get a property from an XMLRPC, it just returns another XMLRPC with the property name tacked onto methodname. */
    public Object get(Object name) {
        return new XMLRPC(url, (methodname.equals("") ? "" : methodname + ".") + name.toString(), http);
    }

    public XMLRPC(String url, String methodname) {
        this(url, methodname, new HTTP(url));
    }

    public XMLRPC(String url, String methodname, HTTP http) {
        this.http = http;
        this.url = url;
        this.methodname = methodname;
    }


    // Helper Classes ///////////////////////////////////////////////////////////////////////////////////

    /** CharArrayWriter that lets us touch its buffer */
    protected static class AccessibleCharArrayWriter extends CharArrayWriter {
        public char[] getBuf() { return buf; }
        public AccessibleCharArrayWriter(int i) { super(i); }
    }

    /** private filter class to make sure that network transfers don't interfere with UI responsiveness */
    private static class Filter extends FilterInputStream {
        public Filter(InputStream is) { super(is); }
        public int read() throws IOException {
            java.lang.Thread.yield();
            while(MessageQueue.nonThreadEventsInQueue > 0) try { java.lang.Thread.sleep(100); } catch (Exception e) { };
            return super.read();
        }
        public int read(byte[] b) throws IOException {
            java.lang.Thread.yield();
            while(MessageQueue.nonThreadEventsInQueue > 0) try { java.lang.Thread.sleep(100); } catch (Exception e) { };
            return super.read(b);
        }
        public int read(byte[] b, int i, int j) throws IOException {
            java.lang.Thread.yield();
            while(MessageQueue.nonThreadEventsInQueue > 0) try { java.lang.Thread.sleep(100); } catch (Exception e) { };
            return super.read(b, i, j);
        }
    }

}
