// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import org.xwt.js.*;
import org.xwt.util.*;
import org.bouncycastle.util.encoders.Base64;

/** Singleton class that provides all functionality in the xwt.* namespace */
public final class XWT extends JS.Obj {

    public static final XWT singleton = new XWT();
    private final JS xwtMath = new XWTMath();
    private final JS xwtString = new XWTString();

    /** each key is a string representing a filename which the user has already given XWT permission to write to */
    private static Hashtable safeFiles = new Hashtable();

    public Object get(Object name) {
        if (name.equals("alt")) return Surface.alt ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("control")) return Surface.control ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("shift")) return Surface.shift ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("clipboard")) return Platform.getClipBoard();
        else if (name.equals("static")) return Static.getStatic("");
        else if (name.equals("button")) {
            if (Surface.button1 && !Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && !Surface.button2 && Surface.button3) return new Integer(1);
            else return new Integer(0);
        }
        else return super.get(name);
    }

    public void put(Object name, Object value) {
        if (name.equals("thread") && value != null && value instanceof JS.Callable) ThreadMessage.newthread((JS.Callable)value);
        else if (name.equals("clipboard")) Platform.setClipBoard(value.toString());
        else if (name.equals("proxyAuthorization")) {
            // FIXME: undocumented, possibly insecure
            Proxy.Authorization.authorization = value.toString();
            Proxy.Authorization.waitingForUser.release();
        } else super.put(name, value);
    }

    private XWT() {
        super.put("maxdim", new Integer(Short.MAX_VALUE));
        super.put("origin", Main.origin);
        super.put("altKeyName", Platform.altKeyName());
        super.put("screenWidth", new Integer(Platform.getScreenWidth()));
        super.put("screenHeight", new Integer(Platform.getScreenHeight()));
        super.put("fileSeparator", File.separator);
        super.put("homeDir", System.getProperty("user.home"));
        super.put("tempDir", System.getProperty("java.io.tempdir"));
        super.put("math", xwtMath);
        super.put("string", xwtString);

        super.put("newBrowserWindow", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() != 1 || args.elementAt(0) == null) return null;
            Platform.newBrowserWindow(args.elementAt(0).toString());
            return null;
        }});

        super.put("parseInt",xwtString.get("parseInt"));
        super.put("parseFloat",xwtString.get("parseFloat"));
        
        super.put("yield", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            sleep(0);
            return null;
        }});

        super.put("theme", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
                if (args.length() != 2) return null;
                if (args.elementAt(0) == null || args.elementAt(1) == null) return null;
                for(int i=1; i<args.length(); i++) {
                    if (args.elementAt(i) instanceof String) {
                        String from = (String)args.elementAt(0);
                        String to = (String)args.elementAt(i);
                        if (Log.on) Log.log(this, "retheming from " + from + " to " + to);
                        Resources.mapFrom.addElement(from);
                        Resources.mapTo.addElement(to);
                    }
                }
                JS.Callable callback = args.elementAt(args.length() - 1) instanceof JS.Callable ?
                    (JS.Callable)args.elementAt(args.length() - 1) : null;
                Template.retheme(callback);
                return null;
        }});
            
        super.put("println", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() != 1) return null;
            if (Log.on) Log.logJS(this, (args.elementAt(0) == null ? "**null**" : args.elementAt(0).toString()));
            return null;
        }});

        super.put("date", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            Log.log(XWT.class, "date not implemented");
            //throw new Error("not implemented");
            return null;
        }});

        super.put("regexp", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            return new Regexp(args);
        }});

        super.put("listfonts", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            Object[] fonts = Platform.listFonts();
            JS.Array ret = new JS.Array();
            for(int i=0; i<fonts.length; i++) ret.addElement(fonts[i]);
            return ret;
        }});

        super.put("xmlrpc", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() != 1 || args.elementAt(0) == null) return null;
            return new XMLRPC(args.elementAt(0).toString(), "");
        }});

        super.put("soap", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() == 1 && args.elementAt(0) != null) return new SOAP(args.elementAt(0).toString(), "", null, null);
            else if (args.length() == 2 && args.elementAt(0) != null && args.elementAt(1) != null)
                return new SOAP(args.elementAt(0).toString(), "", args.elementAt(1).toString(), null);
            else if (args.length() == 3 && args.elementAt(0) != null && args.elementAt(1) != null && args.elementAt(2) != null)
                return new SOAP(args.elementAt(0).toString(), "", args.elementAt(1).toString(), args.elementAt(2).toString());
            else return null;
        }});

        super.put("textwidth", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() < 1 || args.length() > 2) return null;
            if (args.elementAt(0) == null || (args.length() == 2 && args.elementAt(1) == null)) return null;
            String font = args.length() == 1 ? Platform.getDefaultFont() : args.elementAt(0).toString();
            String text = args.length() == 1 ? args.elementAt(0).toString() : args.elementAt(1).toString();
            XWF xwf = XWF.getXWF(font);
            if (xwf == null) return new Integer(Platform.stringWidth(font, text));
            else return new Integer(xwf.stringWidth(text));
        }});

        super.put("textheight", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() > 1) return null;
            if (args.length() == 1 && args.elementAt(0) == null) return null;
            String font = args.length() == 0 || args.elementAt(0) == null ? Platform.getDefaultFont() : args.elementAt(0).toString();
            XWF xwf = XWF.getXWF(font);
            if (xwf == null) return new Integer(Platform.getMaxAscent(font) + Platform.getMaxDescent(font));
            else return new Integer(xwf.getMaxAscent() + xwf.getMaxDescent());
        }});
        
        super.put("newBox", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() > 0) Log.log(XWT.class, "DEPRECATED: xwt.newBox() with multiple arguments is deprecated; use xwt.newBox().apply()");
            JS.Callable callback = null;
            for(int i=1; i<args.length(); i++)
                if (args.elementAt(i) instanceof JS.Callable && callback == null)
                    callback = (JS.Callable)args.elementAt(i);
            Box ret = new Box(args.length() == 0 || args.elementAt(0) == null ? "box" : args.elementAt(0).toString(),
                              Template.defaultImportList, callback);
            for(int i=1; i<args.length(); i++)
                if (args.elementAt(i) instanceof Box)
                    ret.put(ret.numChildren(), (Box)args.elementAt(i));
            for(int i=1; i<args.length(); i++)
                if (args.elementAt(i) instanceof JS && !(args.elementAt(i) instanceof Box) && !(args.elementAt(i) instanceof JS.Callable)) {
                    JS s = (JS)args.elementAt(i);
                    Object[] keys = s.keys();
                    for(int j=0; j<keys.length; j++) ret.put(keys[j].toString(), s.get(keys[j].toString()));
                }
            return ret;
        }});

        super.put("sleep", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args != null && (args.length() != 1 || args.elementAt(0) == null)) return null;
            int i = args == null ? 0 : SpecialBoxProperty.stoi(args.elementAt(0).toString());
            sleep(i);
            return null;
        }});

        super.put("openFile", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() != 1) return null;
            String file = Platform.fileDialog(args.elementAt(0).toString(), false);
            return file == null ? null : new ByteStream(file);
        }});

        super.put("saveFile", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() != 2) return null;
            if (!(args.elementAt(1) instanceof ByteStream)) return null;
            String file = args.elementAt(0).toString();
            if (safeFiles.get(Platform.isCaseSensitive() ? file : file.toLowerCase()) == null) {
                file = Platform.fileDialog(file, true);
                if (file == null) return null;
                safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
            }
            try {
                ((ByteStream)args.elementAt(1)).writeTo(new FileOutputStream(file));
                return null;
            } catch (IOException e) {
                if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                if (Log.on) Log.log(ByteStream.class, e);
                throw new JS.Exn("error while writing a ByteStream to a file");
            }
        }});

        super.put("saveFileAs", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args.length() != 2) return null;
            if (!(args.elementAt(1) instanceof ByteStream)) return null;
            String file = args.elementAt(0).toString();
            file = Platform.fileDialog(file, true);
            if (file == null) return null;
            safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
            try {
                ((ByteStream)args.elementAt(1)).writeTo(new FileOutputStream(file));
                return null;
            } catch (IOException e) {
                if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                if (Log.on) Log.log(ByteStream.class, e);
                throw new JS.Exn("error while writing a ByteStream to a file");
            }
        }});

        super.put("utfEncode", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
            if (args == null || args.length() != 1) return null;
            return new ByteStream(args.elementAt(0).toString().getBytes());
        }});

        super.put("parseHTML", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
                if (args == null || args.length() != 1 || args.elementAt(0) == null) return null;
                try {
                    if (args.elementAt(0) instanceof ByteStream) {
                        return HTML.parseReader(new InputStreamReader(((ByteStream)args.elementAt(0)).getInputStream()));
                    } else {
                        return HTML.parseReader(new StringReader(args.elementAt(0).toString()));
                    }
                } catch (IOException e) {
                    if (Log.on) Log.log(HTML.class, "IO Exception while parsing HTML");
                    if (Log.on) Log.log(HTML.class, e);
                    throw new JS.Exn("error while parsing HTML");
                }
            }
        });
    
    super.put("recursivePrintObject", new JS.Callable() { public Object call(JS.Array args) {
        if (args.length() != 1) return null;
        recurse("", "", args.elementAt(0));
        return null;
    }});

    super.put("loadArchive", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
        if (!ThreadMessage.suspendThread()) return null;
        try {
            if (args == null || args.length() < 1 || args.elementAt(0) == null) return null;
            URL u = new URL(args.elementAt(0).toString());
            
            JS.Callable callback = null;
            if (args.length() == 2 && args.elementAt(1) != null && args.elementAt(1) instanceof JS.Callable)
                callback = (JS.Callable)args.elementAt(1);
            
            if (!u.getFile().endsWith(".xwar")) {
                if (Log.on) Log.log(this, "Error: archive names must end with .xwar: " + u.getFile());
                throw new JS.Exn("Error: archive names must end with .xwar: " + u.getFile());
            }
            
            if (u.getProtocol().equals("http")) {
                HTTP http = new HTTP(u.toString());
                if (Main.originAddr == null) {
                    try {
                        Main.originHost = u.getHost();
                        Main.originAddr = InetAddress.getByName(Main.originHost);
                    } catch (UnknownHostException e) {
                        if (Log.on) Log.log(this, "couldn't resolve " + u.getHost() + "; proceeding without permissions");
                        Main.originAddr = InetAddress.getByName("0.0.0.0");
                    }
                } else {
                    Main.originAddr = InetAddress.getByName("0.0.0.0");
                }
                HTTP.HTTPInputStream in = http.GET();
                Resources.loadArchive(in, in.getContentLength(), callback);
                
            } else if (u.getProtocol().equals("file")) {
                if (Main.originAddr != null) {
                    if (Log.on) Log.log(this, "scripts downloaded from the network may not load xwars from the local filesystem");
                    throw new JS.Exn("scripts downloaded from the network may not load xwars from the local filesystem");
                }
                Resources.loadArchive(new FileInputStream(u.getFile()), (int)new File(u.getFile()).length(), callback);
                
            } else {
                if (Log.on) Log.log(this, "unknown protocol \"" + u.getProtocol() + "\"");
                throw new JS.Exn("unknown protocol \"" + u.getProtocol() + "\"");
            }
            
        } catch (MalformedURLException me) {
            if (Log.on) Log.log(this, "Malformed URL: " + args.elementAt(0));
            if (Log.on) Log.log(this, me);
            throw new JS.Exn(me.toString());
            
        } catch (IOException ioe) {
            if (Log.on) Log.log(this, "IOException while loading archive:");
            if (Log.on) Log.log(this, ioe);
            throw new JS.Exn(ioe.toString());
            
        } finally {
            ThreadMessage.resumeThread();
            
        }
        return null;
    }});

    super.put("prefetchImage", new JS.Callable() { public Object call(JS.Array args) throws JS.Exn {
        if (args == null || args.length() < 1 || args.elementAt(0) == null) return null;
        Box.getImage(args.elementAt(0).toString(),
                     args.length() > 1 && args.elementAt(1) instanceof JS.Callable ? (JS.Callable)args.elementAt(1) : null);
        return null;
    }});
    }

    private static void recurse(String indent, String name, Object o) {
        if (!name.equals("")) name += " : ";

        if (o == null) {
            Log.logJS(indent + name + "<null>");

        } else if (o instanceof JS.Array) {
            Log.logJS(indent + name + "<array>");
            JS.Array na = (JS.Array)o;
            for(int i=0; i<na.length(); i++)
                recurse(indent + "  ", i + "", na.elementAt(i));

        } else if (o instanceof JS) {
            Log.logJS(indent + name + "<object>");
            JS s = (JS)o;
            Object[] keys = s.keys();
            for(int i=0; i<keys.length; i++)
                recurse(indent + "  ", keys[i].toString(),
                        (keys[i] instanceof Integer) ?
                        s.get(((Integer)keys[i])) : s.get(keys[i].toString()));

        } else {
            Log.logJS(indent + name + o);

        }
    }


    public static void sleep(int i) {
        java.lang.Thread thread = java.lang.Thread.currentThread();
        if (!(thread instanceof ThreadMessage)) {
            if (Log.on) Log.log(XWT.class, "cannot sleep() or yield() in the foreground thread");
        } else {
            ThreadMessage mythread = (ThreadMessage)thread;
            mythread.done.release();
            if (i > 0) try { java.lang.Thread.sleep(i); } catch (Exception e) { }
            MessageQueue.add(mythread);
            mythread.go.block();
        }
    }
    
    private static class XWTMath extends JS.Obj {
        public XWTMath() {
            JS gs = new JS.GlobalScope();
            put("isNaN",gs.get("isNaN"));
            put("isFinite",gs.get("isFinite"));
            put("NaN",gs.get("NaN"));
            put("Infinity",gs.get("Infinity"));
            setSeal(true);
        }
        public Object get(Object key) {
            Object ret = super.get(key);
            if(ret == null) ret = JS.Math.get(key);
            return ret;
        }
    }
    private static class XWTString extends JS.Obj {
        public XWTString() {
            JS gs = new JS.GlobalScope();
            put("parseInt",gs.get("parseInt"));
            put("parseFloat",gs.get("parseFloat"));
            put("decodeURI",gs.get("decodeURI"));
            put("decodeURIComponent",gs.get("decodeURIComponent"));
            put("encodeURI",gs.get("encodeURI"));
            put("encodeURIComponent",gs.get("encodeURIComponent"));
            put("escape",gs.get("escape"));
            put("unescape",gs.get("unescape"));
            put("fromCharCode",gs.get("stringFromCharCode"));
            setSeal(true);
        }
    }
}
